# Copyright (c) 2022 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK FDH board interface
'''

import time
import smbus2
from smbus2 import i2c_msg

moduleVersionString = "FDH0.1"

SFA_ADDR = 0x5D

class ModFDH:
	""" This is a class that handles interfacing with the ESDK-FDH board. """
	def __init__(self):
		try:
			self.bus = smbus2.SMBus(1)
		except Exception as e:
			raise e

		self._startPeriodicMeasurement()

	def _startPeriodicMeasurement(self):
		""" Starts the sensor periodic measurement mode. """
		try:
			self.bus.write_i2c_block_data(SFA_ADDR, 0x00, [0x06])
		except Exception as e:
			raise e

	def _readSensorData(self):
		""" Reads all available sensor data

		:return: A list of 9 bytes of sensor data
		:rtype: list

		"""

		try:
			write = i2c_msg.write(SFA_ADDR, [0x03, 0x27])
			read = i2c_msg.read(SFA_ADDR, 9)
			self.bus.i2c_rdwr(write)
			time.sleep(0.005)
			self.bus.i2c_rdwr(write, read)
			return list(read)
		except Exception as e:
			raise e

	def readFormaldehyde(self):
		""" Reads a formaldehyde value from the sensor

		:return: A HCHO reading in ppb
		:rtype: int

		"""
		try:
			data = self._readSensorData()
			hcho = int(((data[0] << 8) + data[1]) / 5.0)
			return hcho
		except Exception as e:
			raise e

	def readSensors(self):
		""" Reads sensors and returns a dictionary containing module version and readings.

		:return: A dictionary containing

		.. code-block:: text

			{
				"fdh":{
					"sensor":"FDH0.1",
					"formaldehyde":25
				}
			}

		:rtype: dict

		"""
		try:
			sensorData = {}
			hcho = self.readFormaldehyde()
			sensorData['sensor'] = moduleVersionString
			sensorData['formaldehyde'] = hcho
			return {'fdh': sensorData}
		except Exception as e:
			raise e


# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK PM2 board interface
'''

import time
import smbus2
from smbus2 import i2c_msg

moduleVersionString = "PM20.2"

SPS_ADDR = 0x69

class ModPM2:
    """ This is a class that handles interfacing with the ESDK-PM2 board. """
    def __init__(self):
        try:
            self.bus = smbus2.SMBus(1)
        except Exception as e:
            raise e

        self.__reset()
        self.__wakeSensor()
        self.startMeasurement()

    def __calculateCrc(self, input):
        """ Calculates CRC for communicating with SPS30 """
        ''' Taken from https://github.com/DFRobot/DFRobot_SGP40/blob/master/Python/raspberrypi/DFRobot_SGP40.py '''
        crc = 0xff
        for i in range(0,2):
            crc = crc^input[i]
            for bit in range(0,8):
                if(crc&0x80):
                    crc = ((crc <<1)^0x31)
                else:
                    crc = (crc<<1)
            crc = crc&0xFF
        return crc

    def __isDataReady(self):
        """ Queries SPS30 to see if there is available data to read

        :return: A boolean value indicating data readiness
        :rtype: bool

        """
        try:
            write = i2c_msg.write(SPS_ADDR, [0x02, 0x02])
            read = i2c_msg.read(SPS_ADDR, 3)
            self.bus.i2c_rdwr(write)
            time.sleep(0.001)
            self.bus.i2c_rdwr(read)
            if list(read)[1] == 0x01:
                return True
            else:
                return False
        except Exception as e:
            raise e

    def __reset(self):
        """ Attempts to reset sensor and delay for the reset time period. """
        try:
            self.bus.write_i2c_block_data(SPS_ADDR, 0xD3, [0x04])
            time.sleep(0.1)
        except Exception as e:
            raise e

    def __wakeSensor(self):
        """ Sends command to wake sensor twice (as per sensor datasheet). """
        try:
            self.bus.write_i2c_block_data(SPS_ADDR, 0x11, [0x03])
            self.bus.write_i2c_block_data(SPS_ADDR, 0x11, [0x03])
        except Exception as e:
            raise e

    def startMeasurement(self):
        """ Starts measurement, configures readings to be unsigned 16-bit integers. """
        try:
            command = [0x00, 0x10, 0x05, 0x00]
            crc = self.__calculateCrc(command[2:4])
            command.append(crc)
            write = i2c_msg.write(SPS_ADDR, command)
            self.bus.i2c_rdwr(write)
        except Exception as e:
            raise e

    def __readSensor(self):
        """ Attempts to read sensor data and pack into a dictionary.

        :return: A dictionary containing:

        .. code-block:: text

            {
                "pm1.0":0,
                "pm2.5":0,
                "pm4.0":0,
                "pm10":0
            }

        :rtype: dict

        """
        try:
            write = i2c_msg.write(SPS_ADDR, [0x03, 0x00])
            read = i2c_msg.read(SPS_ADDR, 30)
            self.bus.i2c_rdwr(write)
            time.sleep(0.001)
            self.bus.i2c_rdwr(read)
            raw_data = list(read)

            # Convert to integer values
            pm10 = (raw_data[0] << 8) + raw_data[1]
            pm25 = (raw_data[3] << 8) + raw_data[4]
            pm40 = (raw_data[6] << 8) + raw_data[7]
            pm100 = (raw_data[9] << 8) + raw_data[10]

            data = {
                "pm1.0": pm10,
                "pm2.5": pm25,
                "pm4.0": pm40,
                "pm10": pm100
            }

            return data
        except Exception as e:
            raise e

    def readSensors(self):
        """ Reads sensors and returns a dictionary containing module version, and all readings.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "pm2":{
                    "sensor":"PM20.2",
                    "pm1.0":0,
                    "pm2.5":0,
                    "pm4.0":0,
                    "pm10":0
                }
            }
            
        Or -1 if data is unavailable

        :rtype: dict, int

        """
        try:
            if self.__isDataReady():
                sensorData = {"sensor": moduleVersionString}
                sensorData.update(self.__readSensor())
                return {'pm': sensorData}
            else:
                return -1
        except Exception as e:
            raise e

    def startFanCleaning(self):
        """ Starts fan cleaning procedure. """
        try:
            self.bus.write_i2c_block_data(SPS_ADDR, 0x56, [0x07])
        except Exception as e:
            raise e


# DFRobot SGP40 VOC index algorithm
# Written by yangfeng(feng.yang@dfrobot.com), 2021.
# Taken from https://github.com/DFRobot/DFRobot_SGP40/tree/master/Python/raspberrypi

VOCALGORITHM_SAMPLING_INTERVAL                           = (1.)
VOCALGORITHM_INITIAL_BLACKOUT                            = (45.)
VOCALGORITHM_VOC_INDEX_GAIN                              = (230.)
VOCALGORITHM_SRAW_STD_INITIAL                            = (50.)
VOCALGORITHM_SRAW_STD_BONUS                              = (220.)
VOCALGORITHM_TAU_MEAN_VARIANCE_HOURS                     = (12.)
VOCALGORITHM_TAU_INITIAL_MEAN                            = (20.)
VOCALGORITHM_INITI_DURATION_MEAN                          = (3600. * 0.75)
VOCALGORITHM_INITI_TRANSITION_MEAN                        = (0.01)
VOCALGORITHM_TAU_INITIAL_VARIANCE                        = (2500.)
VOCALGORITHM_INITI_DURATION_VARIANCE                      = ((3600. * 1.45))
VOCALGORITHM_INITI_TRANSITION_VARIANCE                    = (0.01)
VOCALGORITHM_GATING_THRESHOLD                            = (340.)
VOCALGORITHM_GATING_THRESHOLD_INITIAL                    = (510.)
VOCALGORITHM_GATING_THRESHOLD_TRANSITION                 = (0.09)
VOCALGORITHM_GATING_MAX_DURATION_MINUTES                 = ((60. * 3.))
VOCALGORITHM_GATING_MAX_RATIO                            = (0.3)
VOCALGORITHM_SIGMOID_L                                   = (500.)
VOCALGORITHM_SIGMOID_K                                   = (-0.0065)
VOCALGORITHM_SIGMOID_X0                                  = (213.)
VOCALGORITHM_VOC_INDEX_OFFSET_DEFAULT                    = (100.)
VOCALGORITHM_LP_TAU_FAST                                 = (20.0)
VOCALGORITHM_LP_TAU_SLOW                                 = (500.0)
VOCALGORITHM_LP_ALPHA                                    = (-0.2)
VOCALGORITHM_PERSISTENCE_UPTIME_GAMMA                    = ((3. * 3600.))
VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING      = (64.)
VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__FIX16_MAX          = (32767.)
FIX16_MAXIMUM                                            = 0x7FFFFFFF
FIX16_MINIMUM                                            = 0x80000000
FIX16_OVERFLOW                                           = 0x80000000
FIX16_ONE                                                = 0x00010000

class DFRobot_vocalgorithmParams:
    def __init__(self):
        self.mvoc_index_offset = 0
        self.mtau_mean_variance_hours = 0
        self.mgating_max_duration_minutes = 0
        self.msraw_std_initial=0
        self.muptime=0
        self.msraw=0
        self.mvoc_index=0
        self.m_mean_variance_estimator_gating_max_duration_minutes=0
        self.m_mean_variance_estimator_initialized=0
        self.m_mean_variance_estimator_mean=0
        self.m_mean_variance_estimator_sraw_offset=0
        self.m_mean_variance_estimator_std=0
        self.m_mean_variance_estimator_gamma=0
        self.m_mean_variance_estimator_gamma_initial_mean=0
        self.m_mean_variance_estimator_gamma_initial_variance=0
        self.m_mean_variance_estimator_gamma_mean=0
        self.m_mean_variance_estimator__gamma_variance=0
        self.m_mean_variance_estimator_uptime_gamma=0
        self.m_mean_variance_estimator_uptime_gating=0
        self.m_mean_variance_estimator_gating_duration_minutes=0
        self.m_mean_variance_estimator_sigmoid_l=0
        self.m_mean_variance_estimator_sigmoid_k=0
        self.m_mean_variance_estimator_sigmoid_x0=0
        self.m_mox_model_sraw_mean=0
        self.m_sigmoid_scaled_offset=0
        self.m_adaptive_lowpass_a1=0
        self.m_adaptive_lowpass_a2=0
        self.m_adaptive_lowpass_initialized=0
        self.m_adaptive_lowpass_x1=0
        self.m_adaptive_lowpass_x2=0
        self.m_adaptive_lowpass_x3=0

class DFRobot_VOCAlgorithm:
    
    def __init__(self):
        self.params = DFRobot_vocalgorithmParams()
    def _f16(self,x):
        if x >= 0:
            return int((x)*65536.0 + 0.5)
        else:
            return int((x)*65536.0 - 0.5)

    def _fix16_from_int(self,a):
        return int(a * FIX16_ONE)

    def _fix16_cast_to_int(self,a):
        return int(a) >> 16

    def _fix16_mul(self,inarg0,inarg1):
        inarg0=int(inarg0)
        inarg1=int(inarg1)
        A = (inarg0 >> 16)
        if inarg0<0:
            B = (inarg0&0xFFFFFFFF) & 0xFFFF
        else:
            B = inarg0&0xFFFF
        C = (inarg1 >> 16)
        if inarg1<0:
            D = (inarg1&0xFFFFFFFF) & 0xFFFF
        else:
            D = inarg1&0xFFFF
        AC = (A * C)
        AD_CB = (A * D + C * B)
        BD = (B * D)
        product_hi = (AC + (AD_CB >> 16))
        ad_cb_temp = ((AD_CB) << 16)&0xFFFFFFFF
        product_lo = ((BD + ad_cb_temp))&0xFFFFFFFF
        if product_lo < BD :
            product_hi =product_hi+1
        if ((product_hi >> 31) != (product_hi >>15)):
            return FIX16_OVERFLOW
        product_lo_tmp = product_lo&0xFFFFFFFF
        product_lo = (product_lo - 0x8000)&0xFFFFFFFF
        product_lo = (product_lo-((product_hi&0xFFFFFFFF) >> 31))&0xFFFFFFFF
        if product_lo > product_lo_tmp:
            product_hi = product_hi-1
        result = (product_hi << 16)|(product_lo >> 16)
        result +=1
        return result
    
    def _fix16_div(self,a, b):
        a=int(a)
        b=int(b)
        if b==0 :
            return FIX16_MINIMUM
        if a>=0:
            remainder = a
        else:
            remainder = (a*(-1))&0xFFFFFFFF
        if b >= 0:
            divider = b
        else:
            divider = (b*(-1))&0xFFFFFFFF
        quotient = 0
        bit =0x10000
        while (divider < remainder):
            divider = divider<<1
            bit <<= 1
        if not bit:
            return FIX16_OVERFLOW
        if (divider & 0x80000000):
            if (remainder >= divider):
                quotient |= bit
                remainder -= divider
            divider >>= 1
            bit >>= 1
        while bit and remainder:
            if (remainder >= divider):
                quotient |= bit    
                remainder -= divider
            remainder <<= 1
            bit >>= 1
        if (remainder >= divider):
            quotient+=1
        result = quotient
        if ((a ^ b) & 0x80000000):
            if (result == FIX16_MINIMUM):
                return FIX16_OVERFLOW
            result = -result
        return result
    
    def _fix16_sqrt(self,x):
        x=int(x)
        num=x&0xFFFFFFFF
        result = 0
        bit = 1 << 30
        while (bit > num):
            bit >>=2
        for n in range(0,2):
            while (bit):
                if (num >= result + bit):
                    num = num-(result + bit)&0xFFFFFFFF
                    result = (result >> 1) + bit
                else:
                    result = (result >> 1)
                bit >>= 2
            if n==0:
                if num > 65535:
                    num = (num -result)&0xFFFFFFFF
                    num = ((num << 16) - 0x8000)&0xFFFFFFFF
                    result = ((result << 16) + 0x8000)&0xFFFFFFFF
                else:
                    num = ((num << 16)&0xFFFFFFFF)
                    result =((result << 16)&0xFFFFFFFF)
                bit = 1 << 14
        if (num > result):
                result+=1
        return result
    
    def _fix16_exp(self,x):
        x=int(x)
        exp_pos_values=[self._f16(2.7182818), self._f16(1.1331485), self._f16(1.0157477), self._f16(1.0019550)]
        exp_neg_values=[self._f16(0.3678794), self._f16(0.8824969), self._f16(0.9844964), self._f16(0.9980488)]
        if (x >= self._f16(10.3972)):
            return FIX16_MAXIMUM
        if (x <= self._f16(-11.7835)):
            return 0
        if (x < 0):
            x = -x
            exp_values = exp_neg_values
        else:
            exp_values = exp_pos_values
        res = FIX16_ONE
        arg = FIX16_ONE
        for i in range(0,4):
            while (x >= arg):
                res = self._fix16_mul(res, exp_values[i])
                x -= arg
            arg >>=3
        return res
    
    def vocalgorithm_init(self):
        self.params.mvoc_index_offset = (self._f16(VOCALGORITHM_VOC_INDEX_OFFSET_DEFAULT))
        self.params.mtau_mean_variance_hours = self._f16(VOCALGORITHM_TAU_MEAN_VARIANCE_HOURS)
        self.params.mgating_max_duration_minutes =self._f16(VOCALGORITHM_GATING_MAX_DURATION_MINUTES)
        self.params.msraw_std_initial = self._f16(VOCALGORITHM_SRAW_STD_INITIAL)
        self.params.muptime = self._f16(0.)
        self.params.msraw = self._f16(0.)
        self.params.mvoc_index = 0
        self._vocalgorithm__init_instances()
    
    def _vocalgorithm__init_instances(self):
        self._vocalgorithm__mean_variance_estimator__init()
        self._vocalgorithm__mean_variance_estimator__set_parameters(self._f16(VOCALGORITHM_SRAW_STD_INITIAL), self.params.mtau_mean_variance_hours,self.params.mgating_max_duration_minutes)
        self._vocalgorithm__mox_model__init()
        self._vocalgorithm__mox_model__set_parameters(self._vocalgorithm__mean_variance_estimator__get_std(),self._vocalgorithm__mean_variance_estimator__get_mean())
        self._vocalgorithm__sigmoid_scaled__init()
        self._vocalgorithm__sigmoid_scaled__set_parameters(self.params.mvoc_index_offset)
        self._vocalgorithm__adaptive_lowpass__init()
        self._vocalgorithm__adaptive_lowpass__set_parameters()
    
    def _vocalgorithm_get_states(self,state0,state1):
        state0 = self._vocalgorithm__mean_variance_estimator__get_mean()
        state1 = _vocalgorithm__mean_variance_estimator__get_std()
        return state0,state1
    
    def _vocalgorithm_set_states(self,state0,state1):
        self._vocalgorithm__mean_variance_estimator__set_states(params, state0, state1, self._f16(VOCALGORITHM_PERSISTENCE_UPTIME_GAMMA))
        self.params.msraw = state0
    
    def _vocalgorithm_set_tuning_parameters(self, voc_index_offset, learning_time_hours, gating_max_duration_minutes, std_initial):
        self.params.mvoc_index_offset = self._fix16_from_int(voc_index_offset)
        self.params.mtau_mean_variance_hours = self._fix16_from_int(learning_time_hours)
        self.params.mgating_max_duration_minutes =self._fix16_from_int(gating_max_duration_minutes)
        self.params.msraw_std_initial = self._fix16_from_int(std_initial)
        self._vocalgorithm__init_instances()
    
    def vocalgorithm_process(self, sraw):
        if ((self.params.muptime <= self._f16(VOCALGORITHM_INITIAL_BLACKOUT))):
            self.params.muptime = self.params.muptime + self._f16(VOCALGORITHM_SAMPLING_INTERVAL)
        else:
            if (((sraw > 0) and (sraw < 65000))):
                if ((sraw < 20001)):
                    sraw = 20001
                elif((sraw > 52767)):
                    sraw = 52767
                self.params.msraw = self._fix16_from_int((sraw - 20000))
            self.params.mvoc_index =self._vocalgorithm__mox_model__process(self.params.msraw)
            self.params.mvoc_index =self._vocalgorithm__sigmoid_scaled__process(self.params.mvoc_index)
            self.params.mvoc_index =self._vocalgorithm__adaptive_lowpass__process(self.params.mvoc_index)
            if ((self.params.mvoc_index < self._f16(0.5))):
                self.params.mvoc_index = self._f16(0.5)
            if self.params.msraw > self._f16(0.):
                self._vocalgorithm__mean_variance_estimator__process(self.params.msraw, self.params.mvoc_index)
                self._vocalgorithm__mox_model__set_parameters(self._vocalgorithm__mean_variance_estimator__get_std(),self._vocalgorithm__mean_variance_estimator__get_mean())
        voc_index = self._fix16_cast_to_int((self.params.mvoc_index + self._f16(0.5))) 
        return voc_index
    
    def _vocalgorithm__mean_variance_estimator__init(self):
        self._vocalgorithm__mean_variance_estimator__set_parameters(self._f16(0.),self._f16(0.),self._f16(0.))
        self._vocalgorithm__mean_variance_estimator___init_instances()
    
    def _vocalgorithm__mean_variance_estimator___init_instances(self):
        self._vocalgorithm__mean_variance_estimator___sigmoid__init()
    
    def _vocalgorithm__mean_variance_estimator__set_parameters(self, std_initial, tau_mean_variance_hours, gating_max_duration_minutes):
        self.params.m_mean_variance_estimator_gating_max_duration_minutes = gating_max_duration_minutes
        self.params.m_mean_variance_estimator_initialized = 0
        self.params.m_mean_variance_estimator_mean = self._f16(0.)
        self.params.m_mean_variance_estimator_sraw_offset = self._f16(0.)
        self.params.m_mean_variance_estimator_std = std_initial
        self.params.m_mean_variance_estimator_gamma =self._fix16_div(self._f16((VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING *(VOCALGORITHM_SAMPLING_INTERVAL / 3600.))\
                                                                                  ),(tau_mean_variance_hours +self._f16((VOCALGORITHM_SAMPLING_INTERVAL / 3600.))))
        self.params.m_mean_variance_estimator_gamma_initial_mean =self._f16(((VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING *VOCALGORITHM_SAMPLING_INTERVAL) \
                                                                                /(VOCALGORITHM_TAU_INITIAL_MEAN + VOCALGORITHM_SAMPLING_INTERVAL)))
        self.params.m_mean_variance_estimator_gamma_initial_variance = self._f16(((VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING *VOCALGORITHM_SAMPLING_INTERVAL) \
                                                                                     /(VOCALGORITHM_TAU_INITIAL_VARIANCE + VOCALGORITHM_SAMPLING_INTERVAL)))
        self.params.m_mean_variance_estimator_gamma_mean = self._f16(0.)
        self.params.m_mean_variance_estimator__gamma_variance = self._f16(0.)
        self.params.m_mean_variance_estimator_uptime_gamma = self._f16(0.)
        self.params.m_mean_variance_estimator_uptime_gating = self._f16(0.)
        self.params.m_mean_variance_estimator_gating_duration_minutes = self._f16(0.)
    
    def _vocalgorithm__mean_variance_estimator__set_states(self, mean, std, uptime_gamma):
        self.params.m_mean_variance_estimator_mean = mean
        self.params.m_mean_variance_estimator_std = std
        self.params.m_mean_variance_estimator_uptime_gamma = uptime_gamma
        self.params.m_mean_variance_estimator_initialized = true
        
    
    def _vocalgorithm__mean_variance_estimator__get_std(self):
        return self.params.m_mean_variance_estimator_std
    
    def _vocalgorithm__mean_variance_estimator__get_mean(self):
        return (self.params.m_mean_variance_estimator_mean +self.params.m_mean_variance_estimator_sraw_offset)
    
    def _vocalgorithm__mean_variance_estimator___calculate_gamma(self, voc_index_from_prior):
        uptime_limit = self._f16((VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__FIX16_MAX -VOCALGORITHM_SAMPLING_INTERVAL))
        if self.params.m_mean_variance_estimator_uptime_gamma < uptime_limit:
            self.params.m_mean_variance_estimator_uptime_gamma =(self.params.m_mean_variance_estimator_uptime_gamma +self._f16(VOCALGORITHM_SAMPLING_INTERVAL))
        
        if self.params.m_mean_variance_estimator_uptime_gating < uptime_limit:
            self.params.m_mean_variance_estimator_uptime_gating =(self.params.m_mean_variance_estimator_uptime_gating +self._f16(VOCALGORITHM_SAMPLING_INTERVAL))
        
        self._vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self._f16(1.), self._f16(VOCALGORITHM_INITI_DURATION_MEAN),self._f16(VOCALGORITHM_INITI_TRANSITION_MEAN))
        sigmoid_gamma_mean =self._vocalgorithm__mean_variance_estimator___sigmoid__process(self.params.m_mean_variance_estimator_uptime_gamma)
        gamma_mean =(self.params.m_mean_variance_estimator_gamma +(self._fix16_mul((self.params.m_mean_variance_estimator_gamma_initial_mean -self.params.m_mean_variance_estimator_gamma),sigmoid_gamma_mean)))
        gating_threshold_mean =(self._f16(VOCALGORITHM_GATING_THRESHOLD)\
                                +(self._fix16_mul(self._f16((VOCALGORITHM_GATING_THRESHOLD_INITIAL -VOCALGORITHM_GATING_THRESHOLD)),\
                                 self._vocalgorithm__mean_variance_estimator___sigmoid__process(self.params.m_mean_variance_estimator_uptime_gating))))
        self._vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self._f16(1.),gating_threshold_mean,self._f16(VOCALGORITHM_GATING_THRESHOLD_TRANSITION))
        
        sigmoid_gating_mean =self._vocalgorithm__mean_variance_estimator___sigmoid__process(voc_index_from_prior)
        self.params.m_mean_variance_estimator_gamma_mean =(self._fix16_mul(sigmoid_gating_mean, gamma_mean))
        
        self._vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self._f16(1.), self._f16(VOCALGORITHM_INITI_DURATION_VARIANCE),self._f16(VOCALGORITHM_INITI_TRANSITION_VARIANCE))
        
        sigmoid_gamma_variance =self._vocalgorithm__mean_variance_estimator___sigmoid__process( self.params.m_mean_variance_estimator_uptime_gamma)
        
        gamma_variance =(self.params.m_mean_variance_estimator_gamma +\
                        (self._fix16_mul((self.params.m_mean_variance_estimator_gamma_initial_variance \
                                          -self.params.m_mean_variance_estimator_gamma),\
                                          (sigmoid_gamma_variance - sigmoid_gamma_mean))))
        
        gating_threshold_variance =(self._f16(VOCALGORITHM_GATING_THRESHOLD) \
                                    +(self._fix16_mul(self._f16((VOCALGORITHM_GATING_THRESHOLD_INITIAL -VOCALGORITHM_GATING_THRESHOLD)),\
                                     self._vocalgorithm__mean_variance_estimator___sigmoid__process( self.params.m_mean_variance_estimator_uptime_gating))))
        
        self._vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self._f16(1.), gating_threshold_variance,self._f16(VOCALGORITHM_GATING_THRESHOLD_TRANSITION))
        
        sigmoid_gating_variance =self._vocalgorithm__mean_variance_estimator___sigmoid__process( voc_index_from_prior)
        
        self.params.m_mean_variance_estimator__gamma_variance =(self._fix16_mul(sigmoid_gating_variance, gamma_variance))
        
        self.params.m_mean_variance_estimator_gating_duration_minutes =(self.params.m_mean_variance_estimator_gating_duration_minutes \
                                                                          +(self._fix16_mul(self._f16((VOCALGORITHM_SAMPLING_INTERVAL / 60.)),\
                                                                                           ((self._fix16_mul((self._f16(1.) - sigmoid_gating_mean),\
                                                                                                             self._f16((1. + VOCALGORITHM_GATING_MAX_RATIO))))\
                                                                                              -self._f16(VOCALGORITHM_GATING_MAX_RATIO)))))
        
        if ((self.params.m_mean_variance_estimator_gating_duration_minutes <self._f16(0.))):
            self.params.m_mean_variance_estimator_gating_duration_minutes = self._f16(0.)
        
        if ((self.params.m_mean_variance_estimator_gating_duration_minutes >self.params.m_mean_variance_estimator_gating_max_duration_minutes)):
            self.params.m_mean_variance_estimator_uptime_gating = self._f16(0.)
    
    def _vocalgorithm__mean_variance_estimator__process(self, sraw, voc_index_from_prior):
        if ((self.params.m_mean_variance_estimator_initialized == 0)):
            self.params.m_mean_variance_estimator_initialized = 1
            self.params.m_mean_variance_estimator_sraw_offset = sraw
            self.params.m_mean_variance_estimator_mean = self._f16(0.)
        else:
            if (((self.params.m_mean_variance_estimator_mean >= self._f16(100.)) or (self.params.m_mean_variance_estimator_mean <= self._f16(-100.)))):
                self.params.m_mean_variance_estimator_sraw_offset =(self.params.m_mean_variance_estimator_sraw_offset +self.params.m_mean_variance_estimator_mean)
                self.params.m_mean_variance_estimator_mean = self._f16(0.)
            
            sraw = (sraw - self.params.m_mean_variance_estimator_sraw_offset)
            self._vocalgorithm__mean_variance_estimator___calculate_gamma( voc_index_from_prior)
            delta_sgp = (self._fix16_div((sraw - self.params.m_mean_variance_estimator_mean),self._f16(VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING)))
            if ((delta_sgp < self._f16(0.))):
                c = (self.params.m_mean_variance_estimator_std - delta_sgp)
            else:
                c = (self.params.m_mean_variance_estimator_std + delta_sgp)
            additional_scaling = self._f16(1.)
            if ((c > self._f16(1440.))):
                additional_scaling = self._f16(4.)
            self.params.m_mean_variance_estimator_std = self._fix16_mul(self._fix16_sqrt((self._fix16_mul(additional_scaling,\
                                                                                          (self._f16(VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING) -self.params.m_mean_variance_estimator__gamma_variance)))),\
                                                                          self._fix16_sqrt(((self._fix16_mul(self.params.m_mean_variance_estimator_std,\
                                                                                                            (self._fix16_div(self.params.m_mean_variance_estimator_std,\
                                                                                                            (self._fix16_mul(self._f16(VOCALGORITHM_MEAN_VARIANCE_ESTIMATOR__GAMMA_SCALING),additional_scaling)))))) \
                                                                                          +(self._fix16_mul((self._fix16_div((self._fix16_mul(self.params.m_mean_variance_estimator__gamma_variance,delta_sgp)),additional_scaling))\
                                                                                          ,delta_sgp)))))
            self.params.m_mean_variance_estimator_mean =(self.params.m_mean_variance_estimator_mean +(self._fix16_mul(self.params.m_mean_variance_estimator_gamma_mean,delta_sgp)))
    
    def _vocalgorithm__mean_variance_estimator___sigmoid__init(self):
        self._vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self._f16(0.), self._f16(0.), self._f16(0.))
    
    def _vocalgorithm__mean_variance_estimator___sigmoid__set_parameters(self, L, X0, K):
        self.params.m_mean_variance_estimator_sigmoid_l = L
        self.params.m_mean_variance_estimator_sigmoid_k = K
        self.params.m_mean_variance_estimator_sigmoid_x0 = X0
    
    def _vocalgorithm__mean_variance_estimator___sigmoid__process(self, sample):
        x = (self._fix16_mul(self.params.m_mean_variance_estimator_sigmoid_k,(sample - self.params.m_mean_variance_estimator_sigmoid_x0)))
        if ((x < self._f16(-50.))):
            return self.params.m_mean_variance_estimator_sigmoid_l
        elif ((x > self._f16(50.))):
            return self._f16(0.)
        else:
            return (self._fix16_div(self.params.m_mean_variance_estimator_sigmoid_l,(self._f16(1.) + self._fix16_exp(x))))
    
    def _vocalgorithm__mox_model__init(self):
        self._vocalgorithm__mox_model__set_parameters(self._f16(1.),self._f16(0.))
    
    def _vocalgorithm__mox_model__set_parameters(self,SRAW_STD,SRAW_MEAN):
        self.params.m_mox_model_sraw_std = SRAW_STD
        self.params.m_mox_model_sraw_mean = SRAW_MEAN
    
    def _vocalgorithm__mox_model__process(self,sraw):
        return (self._fix16_mul((self._fix16_div((sraw - self.params.m_mox_model_sraw_mean),(-(self.params.m_mox_model_sraw_std +self._f16(VOCALGORITHM_SRAW_STD_BONUS))))),self._f16(VOCALGORITHM_VOC_INDEX_GAIN)))
    
    def _vocalgorithm__sigmoid_scaled__init(self):
        self._vocalgorithm__sigmoid_scaled__set_parameters(self._f16(0.))
    
    def _vocalgorithm__sigmoid_scaled__set_parameters(self,offset):
        self.params.m_sigmoid_scaled_offset = offset
    
    def _vocalgorithm__sigmoid_scaled__process(self,sample):
        x = (self._fix16_mul(self._f16(VOCALGORITHM_SIGMOID_K),(sample - self._f16(VOCALGORITHM_SIGMOID_X0))))
        if ((x < self._f16(-50.))):
            return self._f16(VOCALGORITHM_SIGMOID_L)
        elif ((x > self._f16(50.))):
            return self._f16(0.)
        else:
            if ((sample >= self._f16(0.))):
                shift = (self._fix16_div((self._f16(VOCALGORITHM_SIGMOID_L) -(self._fix16_mul(self._f16(5.), self.params.m_sigmoid_scaled_offset))),self._f16(4.)))
                return ((self._fix16_div((self._f16(VOCALGORITHM_SIGMOID_L) + shift),(self._f16(1.) + self._fix16_exp(x)))) -shift)
            else:
                return (self._fix16_mul((self._fix16_div(self.params.m_sigmoid_scaled_offset,self._f16(VOCALGORITHM_VOC_INDEX_OFFSET_DEFAULT))),\
                                         (self._fix16_div(self._f16(VOCALGORITHM_SIGMOID_L),(self._f16(1.) + self._fix16_exp(x))))))
    
    def _vocalgorithm__adaptive_lowpass__init(self):
        self._vocalgorithm__adaptive_lowpass__set_parameters()
    
    def _vocalgorithm__adaptive_lowpass__set_parameters(self):
        self.params.m_adaptive_lowpass_a1 =self._f16((VOCALGORITHM_SAMPLING_INTERVAL /(VOCALGORITHM_LP_TAU_FAST + VOCALGORITHM_SAMPLING_INTERVAL)))
        self.params.m_adaptive_lowpass_a2 =self._f16((VOCALGORITHM_SAMPLING_INTERVAL /(VOCALGORITHM_LP_TAU_SLOW + VOCALGORITHM_SAMPLING_INTERVAL)))
        self.params.m_adaptive_lowpass_initialized = 0
    
    def _vocalgorithm__adaptive_lowpass__process(self,sample):
        if ((self.params.m_adaptive_lowpass_initialized == 0)):
            self.params.m_adaptive_lowpass_x1 = sample
            self.params.m_adaptive_lowpass_x2 = sample
            self.params.m_adaptive_lowpass_x3 = sample
            self.params.m_adaptive_lowpass_initialized = 1
        self.params.m_adaptive_lowpass_x1 =((self._fix16_mul((self._f16(1.) - self.params.m_adaptive_lowpass_a1),self.params.m_adaptive_lowpass_x1)) +(self._fix16_mul(self.params.m_adaptive_lowpass_a1, sample)))
        
        self.params.m_adaptive_lowpass_x2 =((self._fix16_mul((self._f16(1.) - self.params.m_adaptive_lowpass_a2),self.params.m_adaptive_lowpass_x2)) +(self._fix16_mul(self.params.m_adaptive_lowpass_a2, sample)))
        
        abs_delta =(self.params.m_adaptive_lowpass_x1 - self.params.m_adaptive_lowpass_x2)
        
        if ((abs_delta < self._f16(0.))):
            abs_delta = (-abs_delta)
        F1 = self._fix16_exp((self._fix16_mul(self._f16(VOCALGORITHM_LP_ALPHA), abs_delta)))
        tau_a =((self._fix16_mul(self._f16((VOCALGORITHM_LP_TAU_SLOW - VOCALGORITHM_LP_TAU_FAST)),F1)) +self._f16(VOCALGORITHM_LP_TAU_FAST))
        a3 = (self._fix16_div(self._f16(VOCALGORITHM_SAMPLING_INTERVAL),(self._f16(VOCALGORITHM_SAMPLING_INTERVAL) + tau_a)))
        self.params.m_adaptive_lowpass_x3 =((self._fix16_mul((self._f16(1.) - a3), self.params.m_adaptive_lowpass_x3)) +(self._fix16_mul(a3, sample)))
        return self.params.m_adaptive_lowpass_x3


# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
Logging functions
'''

import logging

_format = "%(asctime)s [%(levelname)s] %(name)s %(threadName)s: %(message)s"

def getStreamHandler(debug):
    """ Returns a stream handler.

    :param debug: Enables debug logging.
    :type debug: bool
    
    """
    stream_handler = logging.StreamHandler()
    if debug:
        stream_handler.setLevel(logging.DEBUG)
    else:
        stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(_format))
    return stream_handler

def getLogFileHandler(debug, logFile='/aq/log/aq.log'):
    """ Returns a file logging handler with the log level set to DEBUG or INFO.

    :param logFile: Log file path (default is /aq/log/aq.log).
    :type logFile: str, optional
    
    """
    file_handler_log = logging.FileHandler(logFile, mode='w', encoding='utf-8')
    if debug:
        file_handler_log.setLevel(logging.DEBUG)
    else:
        file_handler_log.setLevel(logging.INFO)
    file_handler_log.setFormatter(logging.Formatter(_format))
    return file_handler_log

def getErrorLogFileHandler(errorFile='/aq/log/error.log'):
    """ Returns a file logging handler with the log level set to ERROR.

    :param errorFile: Log file path (default is /aq/log/error.log).
    :type errorFile: str, optional

    """
    file_handler_error = logging.FileHandler(errorFile, mode='w', encoding='utf-8')
    file_handler_error.setLevel(logging.ERROR)
    file_handler_error.setFormatter(logging.Formatter(_format))
    return file_handler_error

def getLogger(name, debug=False, loggingSetup='full'):
    """ Returns a logger object.

    :param debug: Controls whether the debug log level is enabled (default is False)
    :type debug: bool, optional
    :param loggingSetup: One of 'off', 'error' or 'full' log levels (default is 'full')
    :type loggingSetup: str, optional

    """
    logger = logging.getLogger(name)
    if logger.hasHandlers():
        logger.handlers.clear()
    
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.addHandler(getStreamHandler(debug))

    if loggingSetup == 'full':
        logger.debug("Setting file logging to full")
        logger.addHandler(getLogFileHandler(debug))
        logger.addHandler(getErrorLogFileHandler())
    elif loggingSetup == 'error':
        logger.debug("Setting file logging to error only")
        logger.addHandler(getErrorLogFileHandler())
    elif loggingSetup == 'off':
        logger.debug("Setting file logging off")
    else:
        logger.warn("Invalid file logging level: {}".format(loggingSetup))

    return logger


# Copyright (c) 2022 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK NRD board interface
'''

import time
import smbus2
from smbus2 import i2c_msg

moduleVersionString = "NRD0.1"

MOD_ADDR = 0x60

class ModNRD:
	""" This is a class that handles interfacing with the ESDK-NRD board. """
	def __init__(self):
		try:
			self.bus = smbus2.SMBus(1)
		except Exception as e:
			raise e

	def enableI2CWatchdog(self):
		""" Enable NRD I2C watchdog functionality. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x07, [0x01])
		except Exception as e:
			raise e

	def disableI2CWatchdog(self):
		""" Disable NRD I2C watchdog functionality. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x07, [0x00])
		except Exception as e:
			raise e

	def enableEventLed(self):
		""" Enable the event detection LED. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x04, [0x01])
		except Exception as e:
			raise e

	def disableEventLed(self):
		""" Disable the event detection LED. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x04, [0x00])
		except Exception as e:
			raise e

	def enableEventGpio(self):
		""" Enable the event detection GPIO output. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x06, [0x01])
		except Exception as e:
			raise e

	def disableEventGpio(self):
		""" Disable the event detection GPIO output. """
		try:
			self.bus.write_i2c_block_data(MOD_ADDR, 0x06, [0x00])
		except Exception as e:
			raise e

	def getEventLedEnabledState(self):
		""" Get the event LED enabled state.

		:return: A boolean value indicating the event LED enabled status
		:rtype: bool

		"""
		try:
			status = self.bus.read_byte_data(MOD_ADDR, 0x04)
			if status == 0x01:
				return True
			if status == 0x00:
				return False
		except Exception as e:
			raise e

	def getEventGpioEnabledState(self):
		""" Get the event GPIO enabled state.

		:return: A boolean value indicating the event GPIO enabled status
		:rtype: bool

		"""
		try:
			status = self.bus.read_byte_data(MOD_ADDR, 0x06)
			if status == 0x01:
				return True
			if status == 0x00:
				return False
		except Exception as e:
			raise e

	def getI2CWatchdogEnabledState(self):
		""" Get the NRD I2C watchdog enabled state.

		:return: A boolean value indicating the watchdog enabled status
		:rtype: bool

		"""
		try:
			status = self.bus.read_byte_data(MOD_ADDR, 0x07)
			if status == 0x01:
				return True
			if status == 0x00:
				return False
		except Exception as e:
			raise e

	def readCountsPerSecond(self):
		""" Get the current CPS rate.

		:return: An integer representing counts-per-second
		:rtype: int

		"""

		try:
			cps = self.bus.read_byte_data(MOD_ADDR, 0x01)
			return cps
		except Exception as e:
			raise e

	def readCountsPerMinute(self):
		""" Get the current CPM rate.

		:return: An integer representing counts-per-minute
		:rtype: int

		"""

		try:
			write = i2c_msg.write(MOD_ADDR, [0x02])
			read = i2c_msg.read(MOD_ADDR, 2)
			self.bus.i2c_rdwr(write, read)
			data = list(read)

			cpm = int.from_bytes(data, "big")
			return cpm
		except Exception as e:
			raise e

	def readTotalCounts(self):
		""" Get the total accumulated event count

		:return: An integer representing accumulated event counts
		:rtype: int

		"""

		try:
			write = i2c_msg.write(MOD_ADDR, [0x03])
			read = i2c_msg.read(MOD_ADDR, 4)
			self.bus.i2c_rdwr(write, read)
			data = list(read)

			count = int.from_bytes(data, "big")
			return count
		except Exception as e:
			raise e

	def resetCounts(self):
		""" Reset all event counters. """
		try:
			self.bus.write_byte_data(MOD_ADDR, 0x05, 0x01)
		except Exception as e:
			raise e

	def readSensors(self):
		""" Reads sensors and returns a dictionary containing module version and readings.

		:return: A dictionary containing

		.. code-block:: text

			{
				"nrd":{
					"sensor":"NRD0.1",
					"cps":23,
					"cpm":483,
					"totalCounts":9465
				}
			}

		:rtype: dict

		"""

		try:
			sensorData = {}
			cpsReading = self.readCountsPerSecond()
			cpmReading = self.readCountsPerMinute()
			totalReading = self.readTotalCounts()

			sensorData['sensor'] = moduleVersionString
			sensorData['cps'] = cpsReading
			sensorData['cpm'] = cpmReading
			sensorData['totalCounts'] = totalReading

			return {'nrd': sensorData}
		except Exception as e:
			raise e


# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK THV board interface
'''

import smbus2
from smbus2 import i2c_msg
import time
from .DFRobot_SGP40_VOCAlgorithm import DFRobot_VOCAlgorithm

moduleVersionString = "THV0.2"

SHT_ADDR = 0x44
SGP_ADDR = 0x59

# SGP40 commands
SGP40_MEASURE_RAW_SIGNAL = [0x26, 0x0F]
SGP40_EXECUTE_SELF_TEST = [0x28, 0x0E]
SGP40_TURN_HEATER_OFF = [0x36, 0x15]
SGP4X_GET_SERIAL_NUMBER = [0x36, 0x82]
SGP40_MEASURE_RAW_NO_COMP = [0x26, 0x0F, 0x80, 0x00, 0xA2, 0x66, 0x66, 0x93]


class ModTHV:
    """ This is a class that handles interfacing with the ESDK-THV board. """
    def __init__(self):
        self.algorithm = DFRobot_VOCAlgorithm()
        self.algorithm.vocalgorithm_init()

        try:
            self.bus = smbus2.SMBus(1)
        except Exception as e:
            raise e
        
        startTime = int(time.time())
        while (int(time.time()) - startTime < 10):
            self.readVocIndex()

    def __sgp40_crc(self, data1, data2):
        """ Calculates a CRC for two bytes of data, according to SGP40 datasheet. """
        ''' Taken from https://github.com/DFRobot/DFRobot_SGP40/blob/master/Python/raspberrypi/DFRobot_SGP40.py '''
        crc = 0xff
        list = [data1, data2]
        for i in range(0,2):
            crc = crc^list[i]
            for bit in range(0,8):
                if(crc&0x80):
                    crc = ((crc <<1)^0x31)
                else:
                    crc = (crc<<1)
            crc = crc&0xFF
        return crc

    def __readTempAndHumidityRaw(self):
        """ Queries SHT31 and returns a dictionary of raw temperature and humidity values.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "temp":1234,
                "humidity":5678
            }

        :rtype: dict

        """
        try:
            ''' Send high repeatability measurement command without clock stretching '''
            self.bus.write_i2c_block_data(SHT_ADDR, 0x24, [0x00])

            time.sleep(0.5)

            raw_data = self.bus.read_i2c_block_data(SHT_ADDR, 0x00, 6)
            raw_temp = raw_data[0] * 256 + raw_data[1]
            raw_humidity = raw_data[3] * 256 + raw_data[4]

            values = {
                "temp": raw_temp,
                "humidity": raw_humidity
            }

            return values
        except Exception as e:
            raise e

    def readTempAndHumidity(self):
        """ Queries SHT31 and returns a dictionary of temperature and humidity values.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "temperature":12.3,
                "humidity":50.3
            }

        :rtype: dict

        """
        try:
            v = self.__readTempAndHumidityRaw()

            ''' Calculations taken from SHT31 datasheet, section 4.13 '''
            temperature = round(-45 + (175 * v['temp'] / 65535.0), 1)
            humidity = round(100 * v['humidity'] / 65535.0, 1)

            values = {
                "temperature": temperature,
                "humidity": humidity
            }

            return values
        except Exception as e:
            raise e

    def readVocRaw(self):
        """ Returns a compensated raw VOC value. 

        :return: An integer VOC value
        :rtype: int
        """
        try:
            th = self.__readTempAndHumidityRaw()
            ''' Split values into upper and lower bytes to prepare for sending to sensor '''
            tempUpperByte = int(th['temp']) >> 8
            tempLowerByte = int(th['temp']) & 0xFF

            humidityUpperByte = int(th['humidity']) >> 8
            humidityLowerByte = int(th['humidity']) & 0xFF

            tempCrc = self.__sgp40_crc(tempUpperByte, tempLowerByte)
            humidityCrc = self.__sgp40_crc(humidityUpperByte, humidityLowerByte)

            ''' Perform write and then read after 30ms delay, as per datasheet '''
            write = i2c_msg.write(SGP_ADDR, [0x26, 0x0F, humidityUpperByte, humidityLowerByte, humidityCrc, tempUpperByte, tempLowerByte, tempCrc])
            read = i2c_msg.read(SGP_ADDR, 3)
            self.bus.i2c_rdwr(write)
            time.sleep(0.03)
            self.bus.i2c_rdwr(read)
            read = list(read)

            if self.__sgp40_crc(read[0], read[1]) == read[2]:
                vocRaw = (read[0] << 8) + read[1]
                return vocRaw
            else:
                return -1
        except Exception as e:
            raise e

    def readVocIndex(self):
        """ Returns a calculated VOC index value.

        :return: An integer VOC index value, or -1 if unavailable
        :rtype: int
        """
        vocRaw = self.readVocRaw()

        if vocRaw < 0:
            return -1
        else:
            vocIndex = self.algorithm.vocalgorithm_process(vocRaw)
            return vocIndex

    def readSensors(self):
        """ Reads sensors and returns a dictionary containing module version, and all readings.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "thv":{
                    "sensor":"THV0.2",
                    "temperature":21.2,
                    "humidity":50.3,
                    "vocIndex":100
                }
            }
            
        Or -1 if data is unavailable

        :rtype: dict, int

        """
        try:
            sensorData = {}
            vocIndex = self.readVocIndex()
            tempAndHumidity = self.readTempAndHumidity()
            if vocIndex != -1 and tempAndHumidity != -1:
                sensorData['vocIndex'] = vocIndex
                sensorData.update(tempAndHumidity)
                sensorData['sensor'] = moduleVersionString
                return {"thv": sensorData}
            else:
                return -1
        except Exception as e:
            raise e


# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK Main board interface
'''

import smbus2
import toml
import threading
import re
import subprocess
import pkg_resources
import imp
import inspect
import os
import RPi.GPIO as GPIO
from gpsdclient import GPSDClient
from . import AppLogger
from . import MAIN, THV, CO2, PM2, NO2, NRD, FDH

possibleModules = {
    "THV": 0x44, 
    "CO2": 0x62, 
    "PM2": 0x69,
    "NO2": 0x40,
    "NRD": 0x60,
    "FDH": 0x5D
}

moduleTypeDict = {
    'THV': THV,
    'CO2': CO2,
    'PM2': PM2,
    'NO2': NO2,
    "NRD": NRD,
    "FDH": FDH
}

# GPIOs used for board features
SENSOR_3V3_EN = 7
SENSOR_5V_EN = 16
BUZZER_PIN = 26
GPIO_LIST = [SENSOR_3V3_EN, SENSOR_5V_EN, BUZZER_PIN]

strip_unicode = re.compile("([^-_a-zA-Z0-9!@#%&=,/'\";:~`\$\^\*\(\)\+\[\]\.\{\}\|\?\<\>\\]+|[^\s]+)")

class ModMAIN:
    """ This class handles the ESDK mainboard, and it's various features.

    :param config: A dictionary containing configuration data with a minimum of:

    .. code-block:: text

        {
            "esdk":{
                "GPS":False
            }
        }
    
    :type config: dict
    :param debug: Debug logging control, defaults to False
    :type debug: bool, optional
    :param loggingLevel: One of 'off', 'error' or 'full' to control file logging, defaults to 'full'
    :type loggingLevel: str, optional
    :param pluginDir: A string value containing a file path to a plugin directory, defaults to None
    :type pluginDir: str, optional
    """
    def __init__(self, config, debug=False, loggingLevel='full', pluginDir=None):
        self.logger = AppLogger.getLogger(__name__, debug, loggingLevel)
        try:
            self.bus = smbus2.SMBus(1)

            GPIO.setmode(GPIO.BCM)
            GPIO.setup(GPIO_LIST, GPIO.OUT)
            self.buzzer_pwm = GPIO.PWM(BUZZER_PIN, 1000)
        except Exception as e:
            raise e

        self.moduleNames = []
        self.sensorModules = {}
        self.sensorData = {}
        self.configDict = config
        self.location = {}
        self.gpsStatus = {"gpsStatus": {}}
        self.pluginDir = pluginDir
        self.pluginsModuleList = []
        self.plugins = []
        self._parseConfig()

    def _parseConfig(self):
        """ Parse config when mainboard initialised """
        if 'ESDK' in self.configDict:
            if 'gps' in self.configDict['ESDK']:
                if self.configDict['ESDK']['gps'] is not None:
                    if self.configDict['ESDK']['gps'] == True:
                        self.logger.info("GPS is enabled")
                        self.gps = GPSDClient(host="localhost")
                        gpsHandlerThreadHandle = threading.Thread(target=self._gpsHandlerThread, daemon=True)
                        gpsHandlerThreadHandle.name = "gpsHandlerThread"
                        gpsHandlerThreadHandle.start()

    def _gpsHandlerThread(self):
        """ Thread for polling GPS module. """
        self.logger.debug("Started GPS handling thread")
        while True:
            try:
                for result in self.gps.dict_stream():
                    if result["class"] == "TPV":
                        self.location['lat'] = result.get("lat", "n/a")
                        self.location['lon'] = result.get("lon", "n/a")
                        self.logger.debug("GPS location {}".format(self.location))
                        self.gpsStatus['gpsStatus'].update({'mode': result.get("mode", 0)})

                    if result["class"] == "SKY":
                        satellitesList = result.get("satellites", "")
                        satellitesUsedCount = 0

                        for satellite in satellitesList:
                            if satellite['used']:
                                satellitesUsedCount = satellitesUsedCount + 1

                        self.gpsStatus['gpsStatus'].update({'satellitesUsed': satellitesUsedCount})

            except Exception as e:
                self.logger.error("Error getting GPS location, reason {}".format(e))

    def _probeModules(self):
        """ Probes I2C bus to attempt to find sensor modules. """
        self.moduleNames.clear()
        self.logger.debug("Starting module probe")
        for module, addr in possibleModules.items():
            try:
                # ADC used on NO2 board is an annoying edge case, does not seemingly acknowledge 0x0
                if module != "NO2":
                    self.bus.write_byte(addr, 0)
                    self.moduleNames.append(module)
                else:
                    # Instead issue reset command, and check for an acknowledgement
                    self.bus.write_byte(addr, 0x06)
                    self.moduleNames.append(module)
            except Exception as e:
                # Ignore any that fail - the modules aren't present on the bus
                pass
        self.logger.info("Found modules {}".format(self.moduleNames))

    def getLocation(self):
        """ Returns a dictionary containing GPS location, or configuration file location if GPS is disabled.

        :return: A dictionary containing:

        .. code-block:: text

            {
                "lat":0.0,
                "lon":0.0
            }

        :rtype: dict

        """
        if self.configDict['ESDK']['gps'] == False or self.configDict['ESDK']['gps'] is None:
            self.location['lat'] = self.configDict['ESDK']['latitude']
            self.location['lon'] = self.configDict['ESDK']['longitude']
            return self.location

        if self.configDict['ESDK']['gps'] == True:
            if "lat" and "lon" in self.location:
                return self.location
            else:
                return {}

    def getGPSStatus(self):
        """ Returns a dictionary containing GPS status. 

        :return: A dictionary containing:

        .. code-block:: text

            {
                "gpsStatus":{
                    "mode":0,
                    "satellites":13,
                    "satellitesUsed":5
                }
            }

        :rtype: dict

        """
        return self.gpsStatus

    def createModules(self):
        """ Discovers and instantiates module objects for use with ``readAllModules()``. """
        self._probeModules()
        self.logger.debug("Creating module objects")
        for moduleName in self.moduleNames:
            try:
                if moduleName == "THV":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModTHV()

                if moduleName == "CO2":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModCO2()

                if moduleName == "PM2":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModPM2()

                if moduleName == "NO2":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModNO2()

                if moduleName == "NRD":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModNRD()

                if moduleName == "FDH":
                    self.sensorModules[moduleName] = moduleTypeDict[moduleName].ModFDH()
            except Exception as e:
                self.logger.error("Could not create module {}, reason {}".format(moduleName, e))

    def readAllModules(self):
        """ Reads all sensor modules and returns a dictionary containing sensor data. """
        try:
            for name, module in self.sensorModules.items():
                self.logger.debug("Trying to read sensor {}".format(name))
                data = module.readSensors()
                if data != -1:
                    self.sensorData.update(data)
        except Exception as e:
            self.logger.error("Could not read module {}, reason {}".format(name, e))

        # Read loaded plugins
        try:
            for plugin in self.plugins:
                pluginName = plugin.__class__.__name__
                self.logger.debug("Trying to read plugin {}".format(pluginName))
                try:
                    data = plugin.readSensors()
                    if data != -1:
                        self.sensorData.update(data)
                except Exception as e:
                    self.logger.error("Could not read plugin {}, reason {}".format(pluginName, e))
        except Exception as e:
            self.logger.error("Error handling plugins, reason {}".format(e))

        self.logger.debug("Sensor data {}".format(self.sensorData))
        return self.sensorData

    def getSerialNumber(self):
        """ Returns a dictionary containing the Raspberry Pi serial number.

        :return: A dictionary containing:

        .. code-block:: text

            {
                "serialNumber":"RPI0123456789"
            }

        :rtype: dict

        """
        try:
            serialNumber = {}
            with open('/sys/firmware/devicetree/base/serial-number') as f:
                serialNumber['hardwareId'] = "RPI{}".format(strip_unicode.sub('',f.read()))
                self.logger.info("Hardware ID is {}".format(serialNumber['hardwareId']))
            return serialNumber
        except Exception as e:
            self.logger.error("Could not retrieve serial number, reason {}".format(e))
            return -1

    def getModuleVersion(self):
        """ Returns a dictionary containing ESDK module version.

        :return: A dictionary containing:

        .. code-block:: text

            {
                "moduleVerson":"0.0.1"
            }

        :rtype: dict

        """
        return {"moduleVersion": pkg_resources.get_distribution('DesignSpark.ESDK').version}

    def getUndervoltageStatus(self):
        """ Returns a dictionary containing the Raspberry Pi throttle status and code.

        :return: A dictionary containing (throttle_state is optional, and only populated should a nonzero code exist)

        .. code-block:: text

            {
                "throttle_state":{
                    "code":0,
                    "throttle_state":""
                }
            }

        :rtype: dict

        """
        try:
            cmdOutput = subprocess.run(["vcgencmd", "get_throttled"], capture_output=True)
            statusData = cmdOutput.stdout.decode('ascii').strip().strip("throttled=")

            code = int(statusData, 16)
            status = {"code": code}
            response = {"throttle_state": status}

            if statusData == "0x0":
                return response

            statusBits = [[0, "Under_Voltage detected"],
                    [1, "Arm frequency capped"],
                    [2, "Currently throttled"],
                    [3, "Soft temperature limit active"],
                    [16, "Under-voltage has occurred"],
                    [17, "Arm frequency capping has occurred"],
                    [18, "Throttling has occurred"],
                    [19, "Soft temperature limit has occurred"]]

            statusStrings = []

            for x in range(0, len(statusBits)):
                statusBitString = statusBits[x][1]
                if (code & (1 << statusBits[x][0])):
                    statusStrings.append(statusBitString)

            status.update({"status_strings": statusStrings})
            response = {"throttle_state": status}
            return response

        except Exception as e:
            self.logger.error("Could not retrieve undervoltage status, reason {}".format(e))
            return -1

    def setPower(self, vcc3=False, vcc5=False):
        """ Switches 3.3V and 5V sensor power supply rails according to supplied arguments.

        :param vcc3: 3.3V sensor power supply status, defaults to False
        :type vcc3: bool, optional
        :param vcc5: 5V sensor power supply status, defaults to False
        :type vcc5: bool, optional
        """
        try:
            self.logger.debug("Setting sensor power rails, 3V3: {}, 5V: {}".format(vcc3, vcc5))
            GPIO.output(SENSOR_3V3_EN, vcc3)
            GPIO.output(SENSOR_5V_EN, vcc5)
        except Exception as e:
            raise e

    def setBuzzer(self, freq=0):
        """ Sets a PWM frequency on the buzzer output.

        :param freq: Buzzer frequency, 0 stops the buzzer
        :type freq: int, optional
        """
        try:
            if freq > 0:
                self.logger.debug("Setting buzzer frequency to {}".format(freq))
                self.buzzer_pwm.start(50)
                self.buzzer_pwm.ChangeFrequency(freq)

            if freq == 0:
                self.logger.debug("Stopping buzzer")
                self.buzzer_pwm.stop()
        except Exception as e:
            raise e

    def loadPlugins(self):
        """ Attempts to load and instantiate plugins from a specified folder. """
        if self.pluginDir == None:
            cwd = os.getcwd()
            self.pluginFullPath = cwd + "/plugins"
            self.logger.debug("No plugin folder provided, using default")
            self.logger.debug("Current working directory: {}, plugins path: {}".format(cwd, self.pluginFullPath))
        else:
            self.pluginFullPath = self.pluginDir

        # Create a list of available plugin modules
        for filename in os.listdir(self.pluginFullPath):
            modulename, extension = os.path.splitext(filename)
            if extension == '.py':
                file, path, descr = imp.find_module(modulename, [self.pluginFullPath])
                if file:
                    try:
                        self.logger.debug("Found plugin module: {}".format(file.name))
                        module = imp.load_module(modulename, file, path, descr)
                        self.pluginsModuleList.append(module)
                    except Exception as e:
                        self.logger.error("Could not load plugin {}! Reason {}".format(file.name, e))

        # Create a list of instantiated plugin classes
        for pluginModule in self.pluginsModuleList:
            for name, obj in inspect.getmembers(pluginModule):
                if inspect.isclass(obj):
                    self.logger.debug("Created plugin class {}".format(obj))
                    self.plugins.append(obj())

        self.logger.info("Loaded {} plugin(s)".format(len(self.plugins)))



# -*- coding: utf-8 -*-
# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

from setuptools import setup

setup(
    name = 'DesignSpark.ESDK',
    namespace_packages=['DesignSpark'],
    packages = ['DesignSpark.ESDK'],
    version = '23.2.0',
    description = 'DesignSpark ESDK support library',
    author = 'RS Components',
    author_email = 'maint@abopen.com',
    url = 'https://github.com/designsparkrs/DesignSpark.ESDK',
    install_requires = ['smbus2','toml','python-geohash','gpsdclient','paho-mqtt','requests', 'python-snappy', 'RPi.GPIO'],
    license = 'MIT License',
    long_description = open('README.rst').read(),
    long_description_content_type="text/x-rst",
    include_package_data = True,
    keywords = 'raspberry pi raspi designspark esdk',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education',
        'Topic :: System :: Hardware',
        'Topic :: System :: Hardware :: Hardware Drivers'
    ]
)


# Copyright (c) 2021 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK CO2 board interface
'''

import time
import smbus2
from smbus2 import i2c_msg

moduleVersionString = "CO20.2"

SCD_ADDR = 0x62

class ModCO2:
    """ This is a class that handles interfacing with the ESDK-CO2 board. """
    def __init__(self):
        try:
            self.bus = smbus2.SMBus(1)
        except Exception as e:
            raise e

        self.__startPeriodicMeasurement()

    def __startPeriodicMeasurement(self):
        """ Starts periodic measurement mode. """
        try:
            self.bus.write_i2c_block_data(SCD_ADDR, 0x21, [0xB1])
        except Exception as e:
            raise e

    def __isDataReady(self):
        """ Query SCD4x to see if there is available readings

        :return: A boolean value indicating if there is a reading available
        :rtype: bool

        """
        try:
            write = i2c_msg.write(SCD_ADDR, [0xE4, 0xB8])
            read = i2c_msg.read(SCD_ADDR, 3)
            self.bus.i2c_rdwr(write, read)
            state = (list(read)[0] << 8) + list(read)[1]
            if state != 0x8000:
                return True
            else:
                return False
        except Exception as e:
            raise e

    def __readSensorData(self):
        """ Reads all available sensor data

        :return: A list of 9 bytes from the sensor
        :rtype: list

        """
        try:
            write = i2c_msg.write(SCD_ADDR, [0xEC, 0x05])
            read = i2c_msg.read(SCD_ADDR, 9)
            self.bus.i2c_rdwr(write, read)
            return list(read)
        except Exception as e:
            raise e

    def readCO2(self):
        """ Reads a CO2 value from the sensor

        :return: A CO2 reading in ppm, or -1 if the sensor is not ready
        :rtype: int

        """
        try:
            if self.__isDataReady():
                v = self.__readSensorData()
                co2 = (v[0] << 8) + v[1]

                return co2
            else:
                return -1
        except Exception as e:
            raise e

    def readTempAndHumidity(self):
        """ Reads temperature and humidity from the sensor

        :return: A dictionary containing:

        .. code-block:: text

            {
                "temp":12.3,
                "humidity":50.3
            }

        Or -1 if sensor data is unavailable

        :rtype: dict, int
        """
        try:
            if self.__isDataReady():
                v = self.__readSensorData()
                
                temperature = round(-45 + (175 * ((v[3] << 8) + v[4]) / 65535.0), 1)
                humidity = round(100 * ((v[6] << 8) + v[7]) / 65535.0, 1)

                rv = {
                "temp": temperature,
                "humidity": humidity
                }

                return rv
            else:
                return -1
        except Exception as e:
            raise e

    def readSensors(self):
        """ Reads sensors and returns a dictionary containing module version, and all readings.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "co2":{
                    "sensor":"CO20.2",
                    "co2":453
                }
            }
            
        Or -1 if data is unavailable

        :rtype: dict, int

        """
        try:
            sensorData = {}
            co2Reading = self.readCO2()
            sensorData['sensor'] = moduleVersionString
            if co2Reading != -1:
                sensorData['co2'] = co2Reading
                return {'co2': sensorData}
            else:
                return -1
        except Exception as e:
            raise e


# Copyright (c) 2022 RS Components Ltd
# SPDX-License-Identifier: MIT License

'''
ESDK NO2 board interface
'''

import time
import smbus2
import threading
import math
from smbus2 import i2c_msg
from statistics import mode

moduleVersionString = "NO20.1"

ADC_ADDR = 0x40

ADC_REF = 3.000

ADC_UPPER = 0x7FFF

class ModNO2:
	""" This is a class that handles interfacing with the ESDK-PM2 board.

	:param sensitivity: Sensitivity code from barcode on sensor
	:type sensitivity: float
	:param tia_gain: Transimpedance amplifier gain from sensor datasheet
	:type tia_gain: float
	:param voffset: Offset voltage used in gas calculation
	:type voffset: float
	:param movingAverageWindow: Window size for the NO2 moving average
	:type movingAverageWindow: int
	"""
	def __init__(self, sensitivity=-20.86, tia_gain=499, voffset=0, movingAverageWindow=15):
		try:
			self.bus = smbus2.SMBus(1)

			# Sensitivity should be read from the back of the sensing element
			# As this varies from device to device
			self.sensitivity = sensitivity

			# TIA gain is provided by the Spec datasheet, and only changes on sensing element type
			self.tia_gain = tia_gain

			# Voffset figure used in calculation of gas concentration
			self.voffset = voffset

			self.movingAverageWindow = movingAverageWindow

			self.no2AverageList = []
			# NO2 value updated in thread
			self.no2Value = 0
		except Exception as e:
			raise e

		self._resetADC()
		adcPollingThreadHandle = threading.Thread(target=self._adcPollingThread, daemon=True)
		adcPollingThreadHandle.name = "no2_adcPollingThreadHandle"
		adcPollingThreadHandle.start()

	def _resetADC(self):
		""" Issues reset command to ADC. """
		try:
			self.bus.write_byte(ADC_ADDR, 0x06)
		except Exception as e:
			raise e

	def _isDataReady(self):
		""" Queries ADC to see if data is available to be read.

		:return: True or false if ADC data is available.
		:rtype: bool

		"""
		try:
			write = i2c_msg.write(ADC_ADDR, [0x24])
			read = i2c_msg.read(ADC_ADDR, 1)
			self.bus.i2c_rdwr(write, read)

			ready = list(read)[0] >> 7

			if ready == 1:
				return True
			else:
				return False
		except Exception as e:
			raise e

	def _readVrefChannel(self):
		""" Read and convert AIN1 (sensor Vref) to a voltage. 

		:return: The voltage of AIN1 (sensor Vref).
		:rtype: float

		"""
		try:
			# 0x81 is the config register data
			# AINp = AIN1, AINn = GND
			# Gain = 1
			# 20 SPS
			# Single conversion
			# External reference
			self.bus.write_byte_data(ADC_ADDR, 0x40, 0x81)

			# Trigger read
			self.bus.write_byte(ADC_ADDR, 0x08)

			time.sleep(0.05)

			if self._isDataReady():
				write = i2c_msg.write(ADC_ADDR, [0x10])
				read = i2c_msg.read(ADC_ADDR, 2)
				self.bus.i2c_rdwr(write, read)

				read = list(read)
				voltage = ((read[0] << 8) + read[1]) * (ADC_REF / ADC_UPPER)
				return round(voltage, 3)
			else:
				return -1
		except Exception as e:
			raise e

	def _readVgasChannel(self):
		""" Read and convert AIN0 (sensor Vgas) to a voltage. 

		:return: The voltage of AIN0 (sensor Vgas).
		:rtype: float

		"""
		try:
			# 0x61 is the config register data
			# AINp = AIN0, AINn = GND
			# Gain = 1
			# 20 SPS
			# Single conversion
			# External reference
			self.bus.write_byte_data(ADC_ADDR, 0x40, 0x61)

			# Trigger read
			self.bus.write_byte(ADC_ADDR, 0x08)

			time.sleep(0.05)

			if self._isDataReady():
				write = i2c_msg.write(ADC_ADDR, [0x10])
				read = i2c_msg.read(ADC_ADDR, 2)
				self.bus.i2c_rdwr(write, read)

				read = list(read)
				voltage = ((read[0] << 8) + read[1]) * (ADC_REF / ADC_UPPER)
				return round(voltage, 3)
			else:
				return -1
		except Exception as e:
			raise e

	def _readVtempChannel(self):
		""" Read and convert AIN2 (sensor Vtemp) to a voltage.

		:return: The voltage of AIN2 (sensor Vtemp).
		:rtype: float
		
		"""
		try:
			# 0xA1 is the config register data
			# AINp = AIN2, AINn = GND
			# Gain = 1
			# 20 SPS
			# Single conversion
			# External reference
			self.bus.write_byte_data(ADC_ADDR, 0x40, 0xA1)

			# Trigger read
			self.bus.write_byte(ADC_ADDR, 0x08)

			time.sleep(0.05)

			if self._isDataReady():
				write = i2c_msg.write(ADC_ADDR, [0x10])
				read = i2c_msg.read(ADC_ADDR, 2)
				self.bus.i2c_rdwr(write, read)

				read = list(read)
				voltage = ((read[0] << 8) + read[1]) * (ADC_REF / ADC_UPPER)
				return round(voltage, 3)
			else:
				return -1
		except Exception as e:
			raise e

	def _calculateMovingAverage(self, new_data_point, data_list, average_size):
		""" Function to calculate a simple moving average """
		# Add fresh data
		data_list.insert(0, new_data_point)
		try:
			# Remove nth item in list
			data_list.pop(average_size)
		except Exception as e:
			pass
		total = 0
		total = sum(data_list)
        
		#value = math.ceil(total / len(data_list))
		value = round(total / len(data_list), 2)
		if(value == None):
			return 0
		else:
			return value

	def _adcPollingThread(self):
		""" Thread that polls the ADC to provid an updated NO2 value every five seconds """
		vref = -1
		vgasList = []
		vgasLastMode = 0

		while vref == -1:
			vref = self._readVrefChannel()

		while True:
			try:
				# Take 10 voltage readings from ADC
				for i in range(1, 10):
					vgas = -1
					# Wait until ADC says data is available
					while vgas == -1:
						vgas = self._readVgasChannel()
						time.sleep(0.1)

					# Append each sample to list and wait before taking another reading
					vgasList.append(vgas)
					time.sleep(0.25)
				# Take mode value for use in calculations to help reduce sensor noise (ADC and inputs seem noisy)
				try:
					vgasLastMode = vgas
					vgas = mode(vgasList)
				except Exception as e:
					vgasList.clear()

				if vgas != -1 and vref != -1:
					vgas0 = vref + self.voffset
					conc = (vgas - vgas0) / (self.tia_gain * 1e3) / (self.sensitivity * 1e-9)
					conc = round(conc, 2)
			except Exception as e:
				pass

			# Calculate moving average to use for value
			self.no2Value = self._calculateMovingAverage(conc, self.no2AverageList, self.movingAverageWindow)

			vgasList.clear()
			time.sleep(2.5)

	def readNO2(self):
		""" Returns an NO2 reading. 

		:return: The NO2 concentration.
		:rtype: float

		"""

		return self.no2Value

	def readSensors(self):
		""" Reads sensor and returns a dictionary containing module version, and all readings.

        :return: A dictionary containing

        .. code-block:: text
            
            {
                "no2":{
                    "sensor":"NO20.1",
                    "no2":2.1,
                }
            }
            
        Or -1 if data is unavailable

        :rtype: dict, int

        """
		try:
			sensorData = {}
			no2 = self.readNO2()
			sensorData['sensor'] = moduleVersionString
			if no2 != -1:
				sensorData['no2'] = no2
				return {'no2': sensorData}
			else:
				return -1
		except Exception as e:
			raise e